@@ -23,6 +23,7 @@
 #include <pcre.h>
 #include <time.h>
 #include <errno.h>
+#include <uuid/uuid.h>
 
 #include "filter-datatypes.h"
 #include "filter.h"
@@ -38,73 +39,6 @@ typedef struct _fs_date_fields {
     int day;
 } fs_date_fields;
 
-
-static char fs_bit_mask_set(short nbit,short value) {
-    
-    if (value)
-        return  0xFF;
-
-    if (nbit == 0) 
-        return 0x7F;
-    else if (nbit == 1)
-        return 0xBF;
-    else if (nbit == 2) 
-        return 0xDF;
-    else if (nbit == 3) 
-        return 0xEF;
-    else if (nbit == 4) 
-        return 0xF7;
-    else if (nbit == 5) 
-        return 0xFB;
-    else if (nbit == 6) 
-        return 0xFD;
-    else if (nbit == 7) 
-        return 0xFE;
-    return 0xFF;
-}
-
-
-static char fs_bit_mask_get(short nbit) {
-    if (nbit == 0) 
-        return 0x80;
-    else if (nbit == 1)
-        return 0x40;
-    else if (nbit == 2) 
-        return 0x20;
-    else if (nbit == 3) 
-        return 0x10;
-    else if (nbit == 4) 
-        return 0x08;
-    else if (nbit == 5) 
-        return 0x04;
-    else if (nbit == 6) 
-        return 0x02;
-    else if (nbit == 7) 
-        return 0x01;
-    return 0x00;
-}
-
-unsigned char *fs_new_bit_array(long n) {
-    int extra = (n % 8) ? 1 : 0;
-    size_t ms = ((n / 8)+extra) * sizeof(unsigned char);
-    unsigned char *p = malloc(ms);
-    memset(p,0xFF,ms);
-    return p;
-}
-
-void fs_bit_array_set(unsigned char *p,long i,int value) {
-    p[i/8] &= fs_bit_mask_set(i % 8,value);
-}
-
-short fs_bit_array_get(unsigned char *p,long i) {
-    return p[i/8] & fs_bit_mask_get(i % 8) ? 1 : 0;
-}
-
-void fs_bit_array_destroy(unsigned char *p) {
-    if (p)
-        free(p);
-}
-
 static fs_value cast_double(fs_value a)
 {
     if (a.valid & fs_valid_bit(FS_V_FP)) {
@@ -486,18 +420,18 @@ fs_value fn_numeric_round(fs_query *q, fs_value a)
 fs_value fn_numeric_add(fs_query *q, fs_value a, fs_value b)
 {
 #if 0
-fs_value_print(a);
-printf(" + ");
-fs_value_print(b);
-printf("\n");
+    fs_value_print(a);
+    printf(" + ");
+    fs_value_print(b);
+    printf("\n");
 #endif
     a = fs_value_promote(q, a, b);
     b = fs_value_promote(q, b, a);
 #if 0
-fs_value_print(a);
-printf(" P+ ");
-fs_value_print(b);
-printf("\n");
+    fs_value_print(a);
+    printf(" P+ ");
+    fs_value_print(b);
+    printf("\n");
 #endif
 
     if (a.attr == b.attr && a.attr != FS_RID_NULL && a.attr != fs_c.empty) {
@@ -592,10 +526,10 @@ fs_value fn_numeric_divide(fs_query *q, fs_value a, fs_value b)
 fs_value fn_equal(fs_query *q, fs_value a, fs_value b)
 {
 #if 0
-fs_value_print(a);
-printf(" = ");
-fs_value_print(b);
-printf("\n");
+    fs_value_print(a);
+    printf(" = ");
+    fs_value_print(b);
+    printf("\n");
 #endif
     if (a.valid & fs_valid_bit(FS_V_TYPE_ERROR)) {
 	return a;
@@ -626,10 +560,10 @@ printf("\n");
 fs_value fn_not_equal(fs_query *q, fs_value a, fs_value b)
 {
 #if 0
-fs_value_print(a);
-printf(" != ");
-fs_value_print(b);
-printf("\n");
+    fs_value_print(a);
+    printf(" != ");
+    fs_value_print(b);
+    printf("\n");
 #endif
     if (a.valid & fs_valid_bit(FS_V_TYPE_ERROR)) {
 	return a;
@@ -699,10 +633,10 @@ fs_value fn_less_than(fs_query *q, fs_value a, fs_value b)
 	return b;
     }
 #if 0
-fs_value_print(a);
-printf(" < ");
-fs_value_print(b);
-printf("\n");
+    fs_value_print(a);
+    printf(" < ");
+    fs_value_print(b);
+    printf("\n");
 #endif
 
     if (a.attr == fs_c.xsd_datetime && b.attr == fs_c.xsd_datetime) return fn_datetime_less_than(q, a, b);
@@ -751,10 +685,10 @@ fs_value fn_greater_than_equal(fs_query *q, fs_value a, fs_value b)
 fs_value fn_less_than_equal(fs_query *q, fs_value a, fs_value b)
 {
 #if 0
-fs_value_print(a);
-printf(" <= ");
-fs_value_print(b);
-printf("\n");
+    fs_value_print(a);
+    printf(" <= ");
+    fs_value_print(b);
+    printf("\n");
 #endif
     if (a.valid & fs_valid_bit(FS_V_TYPE_ERROR)) {
 	return a;
@@ -786,10 +720,10 @@ printf("\n");
 fs_value fn_numeric_equal(fs_query *q, fs_value a, fs_value b)
 {
 #if 0
-fs_value_print(a);
-printf(" ");
-fs_value_print(b);
-printf("\n");
+    fs_value_print(a);
+    printf(" ");
+    fs_value_print(b);
+    printf("\n");
 #endif
     a = fs_value_promote(q, a, b);
     b = fs_value_promote(q, b, a);
@@ -817,10 +751,10 @@ fs_value fn_numeric_less_than(fs_query *q, fs_value a, fs_value b)
     a = fs_value_promote(q, a, b);
     b = fs_value_promote(q, b, a);
 #if 0
-fs_value_print(a);
-printf(" < ");
-fs_value_print(b);
-printf("\n");
+    fs_value_print(a);
+    printf(" < ");
+    fs_value_print(b);
+    printf("\n");
 #endif
 
     if (a.attr == b.attr && a.attr != FS_RID_NULL && a.attr != fs_c.empty) {
@@ -912,10 +846,10 @@ fs_value fn_datetime_equal(fs_query *q, fs_value a, fs_value b)
 fs_value fn_datetime_less_than(fs_query *q, fs_value a, fs_value b)
 {
 #if 0
-fs_value_print(a);
-printf(" < ");
-fs_value_print(b);
-printf(" [dT]\n");
+    fs_value_print(a);
+    printf(" < ");
+    fs_value_print(b);
+    printf(" [dT]\n");
 #endif
     if (a.attr == fs_c.xsd_datetime && b.attr == fs_c.xsd_datetime &&
         (a.in != -1 || b.in != -1))
@@ -951,10 +885,10 @@ fs_value fn_compare(fs_query *q, fs_value a, fs_value b)
     }
 
 #if 0
-fs_value_print(a);
-printf(" <=> ");
-fs_value_print(b);
-printf("\n");
+    fs_value_print(a);
+    printf(" <=> ");
+    fs_value_print(b);
+    printf("\n");
 #endif
     if ((FS_IS_LITERAL(a.attr) && FS_IS_LITERAL(b.attr)) ||
 	(a.attr == fs_c.empty && b.attr == fs_c.empty)) {
@@ -1027,9 +961,9 @@ fs_value fn_logical_or(fs_query *q, fs_value a, fs_value b)
 fs_value fn_not(fs_query *q, fs_value a)
 {
 #if 0
-printf("! ");
-fs_value_print(a);
-printf("\n");
+    printf("! ");
+    fs_value_print(a);
+    printf("\n");
 #endif
     if (fs_is_error(a)) {
 	return a;
@@ -1238,9 +1172,9 @@ fs_value fn_lang(fs_query *q, fs_value a)
 fs_value fn_datatype(fs_query *q, fs_value a)
 {
 #if 0
-printf("datatype(");
-fs_value_print(a);
-printf(")\n");
+    printf("datatype(");
+    fs_value_print(a);
+    printf(")\n");
 #endif
     if (a.valid & fs_valid_bit(FS_V_TYPE_ERROR)) {
 	return a;
@@ -1288,39 +1222,11 @@ printf(")\n");
     return fs_value_uri("error:unresloved");
 }
 
-fs_value fn_matches(fs_query *q, fs_value str, fs_value pat, fs_value flags)
+static int regex_flags(const char *flags)
 {
-    if (str.valid & fs_valid_bit(FS_V_TYPE_ERROR)) {
-	return str;
-    }
-    if (pat.valid & fs_valid_bit(FS_V_TYPE_ERROR)) {
-	return pat;
-    }
-    if (flags.valid & fs_valid_bit(FS_V_TYPE_ERROR)) {
-	return flags;
-    }
-
-    if (!str.lex || !pat.lex) {
-	return fs_value_error(FS_ERROR_INVALID_TYPE,
-                              "argument to fn:matches has no lexical value");
-    }
-
-    if (str.valid & fs_valid_bit(FS_V_RID) && FS_IS_URI(str.rid)) {
-	return fs_value_error(FS_ERROR_INVALID_TYPE, NULL);
-    }
-#if 0
-printf("REGEX ");
-fs_value_print(str);
-printf(", ");
-fs_value_print(pat);
-printf(", ");
-fs_value_print(flags);
-printf("\n");
-#endif
-
     int reflags = PCRE_UTF8;
-    if (flags.lex) {
-	for (char *c = flags.lex; *c; c++) {
+    if (flags) {
+	for (const char *c = flags; *c; c++) {
 	    switch (*c) {
 		case 's':
 		    reflags |= PCRE_DOTALL;
@@ -1336,11 +1242,52 @@ printf("\n");
 		    break;
 		default:
 		    fs_error(LOG_ERR, "unknown regex flag '%c'", *c);
-		    return fs_value_error(FS_ERROR_INVALID_TYPE, "unrecognised flag in fn:matches");
+
+                    return 0;
 	    }
 	}
     }
 
+    return reflags;
+}
+
+fs_value fn_matches(fs_query *q, fs_value str, fs_value pat, fs_value flags)
+{
+    if (str.valid & fs_valid_bit(FS_V_TYPE_ERROR)) {
+	return str;
+    }
+    if (pat.valid & fs_valid_bit(FS_V_TYPE_ERROR)) {
+	return pat;
+    }
+    if (flags.valid & fs_valid_bit(FS_V_TYPE_ERROR)) {
+	return flags;
+    }
+
+    str = fs_value_fill_lexical(q, str);
+    pat = fs_value_fill_lexical(q, pat);
+    if (!str.lex || !pat.lex) {
+	return fs_value_error(FS_ERROR_INVALID_TYPE,
+                              "argument to fn:matches has no lexical value");
+    }
+
+    if (str.valid & fs_valid_bit(FS_V_RID) && FS_IS_URI(str.rid)) {
+	return fs_value_error(FS_ERROR_INVALID_TYPE, NULL);
+    }
+#if 0
+    printf("REGEX ");
+    fs_value_print(str);
+    printf(", ");
+    fs_value_print(pat);
+    printf(", ");
+    fs_value_print(flags);
+    printf("\n");
+#endif
+
+    int reflags = regex_flags(flags.lex);
+    if (!reflags) {
+        return fs_value_error(FS_ERROR_INVALID_TYPE, "unrecognised flag in fn:matches");
+    }
+
     const char *error;
     int erroroffset;
     pcre *re = pcre_compile(pat.lex, reflags, &error, &erroroffset, NULL);
@@ -1383,11 +1330,11 @@ fs_value fn_cast_intl(fs_query *q, fs_value v, fs_rid dt)
 fs_value fn_cast(fs_query *q, fs_value v, fs_value d)
 {
 #if 0
-printf("CAST ");
-fs_value_print(v);
-printf(" -> ");
-fs_value_print(d);
-printf("\n");
+    printf("CAST ");
+    fs_value_print(v);
+    printf(" -> ");
+    fs_value_print(d);
+    printf("\n");
 #endif
     if (FS_IS_URI(d.rid) && FS_IS_LITERAL(v.rid)) {
 	return fn_cast_intl(q, v, d.rid);
@@ -1625,7 +1572,7 @@ fs_value fn_timezone(fs_query *q, fs_value v)
 
 fs_value fn_strstarts(fs_query *q, fs_value arg1, fs_value arg2)
 {
-    if (!fs_is_plain_or_string(arg1) || !fs_is_plain_or_string(arg2)) {
+    if (!fs_arg_compatible(arg1, arg2)) {
         return fs_value_error(FS_ERROR_INVALID_TYPE, NULL);
     }
     arg1 = fs_value_fill_lexical(q, arg1);
@@ -1636,7 +1583,7 @@ fs_value fn_strstarts(fs_query *q, fs_value arg1, fs_value arg2)
 
 fs_value fn_strends(fs_query *q, fs_value arg1, fs_value arg2)
 {
-    if (!fs_is_plain_or_string(arg1) || !fs_is_plain_or_string(arg2)) {
+    if (!fs_arg_compatible(arg1, arg2)) {
         return fs_value_error(FS_ERROR_INVALID_TYPE, NULL);
     }
     arg1 = fs_value_fill_lexical(q, arg1);
@@ -1654,7 +1601,7 @@ fs_value fn_strends(fs_query *q, fs_value arg1, fs_value arg2)
 
 fs_value fn_contains(fs_query *q, fs_value arg1, fs_value arg2)
 {
-    if (!fs_is_plain_or_string(arg1) || !fs_is_plain_or_string(arg2)) {
+    if (!fs_arg_compatible(arg1, arg2)) {
         return fs_value_error(FS_ERROR_INVALID_TYPE, NULL);
     }
     arg1 = fs_value_fill_lexical(q, arg1);
@@ -1663,6 +1610,58 @@ fs_value fn_contains(fs_query *q, fs_value arg1, fs_value arg2)
     return fs_value_boolean(strstr(arg1.lex, arg2.lex) != NULL);
 }
 
+fs_value fn_strbefore(fs_query *q, fs_value arg1, fs_value arg2)
+{
+    if (!fs_arg_compatible(arg1, arg2)) {
+        return fs_value_error(FS_ERROR_INVALID_TYPE, NULL);
+    }
+
+    arg1 = fs_value_fill_lexical(q, arg1);
+    arg2 = fs_value_fill_lexical(q, arg2);
+
+    char *pos = strstr(arg1.lex, arg2.lex);
+    if (!pos) {
+        fs_value ret = fs_value_plain("");
+        ret.attr = arg1.attr;
+
+        return ret;
+    }
+
+    char *new = g_strndup(arg1.lex, pos-arg1.lex);
+    fs_query_add_freeable(q, new);
+
+    fs_value ret = fs_value_plain(new);
+    ret.attr = arg1.attr;
+
+    return ret;
+}
+
+fs_value fn_strafter(fs_query *q, fs_value arg1, fs_value arg2)
+{
+    if (!fs_arg_compatible(arg1, arg2)) {
+        return fs_value_error(FS_ERROR_INVALID_TYPE, NULL);
+    }
+
+    arg1 = fs_value_fill_lexical(q, arg1);
+    arg2 = fs_value_fill_lexical(q, arg2);
+
+    char *pos = strstr(arg1.lex, arg2.lex);
+    if (!pos) {
+        fs_value ret = fs_value_plain("");
+        ret.attr = arg1.attr;
+
+        return ret;
+    }
+
+    char *new = g_strdup(pos + strlen(arg2.lex));
+    fs_query_add_freeable(q, new);
+
+    fs_value ret = fs_value_plain(new);
+    ret.attr = arg1.attr;
+
+    return ret;
+}
+
 fs_value fn_rand(fs_query *q)
 {
     static int seeded = 0;
@@ -1723,4 +1722,28 @@ fs_value fn_sha256(fs_query *q, fs_value arg)
 #endif
 }
 
+fs_value fn_uuid(fs_query *q)
+{
+    uuid_t uu;
+    uuid_string_t uus;
+    uuid_generate(uu);
+    uuid_unparse(uu, uus);
+    char *str = g_strdup_printf("urn:uuid:%s", uus);
+    fs_query_add_freeable(q, str);
+
+    return fs_value_uri(str);
+}
+
+fs_value fn_struuid(fs_query *q)
+{
+    uuid_t uu;
+    uuid_string_t uus;
+    uuid_generate(uu);
+    uuid_unparse(uu, uus);
+    char *str = g_strdup(uus);
+    fs_query_add_freeable(q, str);
+
+    return fs_value_plain(str);
+}
+
 /* vi:set expandtab sts=4 sw=4: */
