@@ -25,10 +25,19 @@ struct _fs_query_state {
     raptor_world *raptor_world;
 
     int verbosity;
+    int cache_stats;
+
+    /* the following cache stats are filled only if verbosity > 0  
+    or cache_stats option in httpd */
+    /* bind stats */
+    unsigned int bind_hits;
+    unsigned int bind_cache_success;
+
     /* the following cache stats are filled only if verbosity > 0 */
     unsigned int cache_hits; /* total queries to the cache */
     unsigned int cache_success_l1; /* number of l1 success hits */
     unsigned int cache_success_l2;  /* number of l2 success hits */
+    double avg_cache_saves_l2; /* avg number of resolve saves in cache l2 */
     unsigned int cache_fail;  /* number of cache hits with no data on l1 or l2 */
     unsigned int pre_cache_total; /* number of items pre cached for the query */
     unsigned int resolve_all_calls; /* total num of resolve_all calls */
@@ -36,6 +45,12 @@ struct _fs_query_state {
     double resolve_unique_elapse; /* total sum of elapsed time on resolve(single rid) calls */
 };
 
+struct _fs_bind_expression {
+    rasqal_variable *var;
+    rasqal_expression *expr;
+    struct _fs_bind_expression *next;
+};
+
 struct _fs_query {
     fs_query_state *qs;
     fsp_link *link;
@@ -53,6 +68,7 @@ struct _fs_query {
     int limit;				/* a user specified limit, or -1 */
     int soft_limit;			/* a limit chosen by the system
 					   to prevent complexity explostion */
+    fs_rid apikey_rid; /* api key rid used in access control for graphs */
     int offset;
     int opt_level;			/* optimisation level in [0,3] */
     int boolean;			/* true if the query succeeded */
@@ -73,6 +89,7 @@ struct _fs_query {
     int parent_block[FS_MAX_BLOCKS];
     int union_group[FS_MAX_BLOCKS];
     raptor_sequence *constraints[FS_MAX_BLOCKS];
+    fs_bind_expression *binds[FS_MAX_BLOCKS];
     int flags;
     fs_rid_vector **pending;
     rasqal_query *rq;
@@ -88,11 +105,13 @@ struct _fs_query {
     fs_rid_vector *default_graphs;
     int console;			/* true if the query is being used from a console app */
     int aggregate;			/* true if the query uses aggregates */
+    int offset_aggregate;   /* offset to be evaluated in result generation */
     long group_length;			/* number of rows in the current group */
     uint64_t *group_rows;		/* row numbers of the rows in the current group */
     unsigned char *apply_constraints; /* bit array initialized to 1s, 
                                         position x shifts to 0 if no apply cons */
     int group_by;
+    GHashTable *tmp_resources;
 };
 
 #endif
