@@ -39,11 +39,10 @@
 #include "import.h"
 #include "debug.h"
 #include "../common/params.h"
+#include "../common/4s-internals.h"
 #include "../common/error.h"
 #include "../common/rdf-constants.h"
 
-//#define DEBUG_BIND
-
 #define DESC_SIZE 1024
 
 #define DEBUG_SIZE(n, thing) printf("@@ %d * sizeof(%s) = %zd\n", n, #thing, n * sizeof(thing))
@@ -298,10 +297,10 @@ static void tree_compact(fs_query *q)
             int parent = q->parent_block[block];
             int mergable = 0;
             if (q->join_type[block] == FS_INNER && q->blocks[block].length > 0) {
-                if (q->constraints[block] == NULL) {
+                if (q->constraints[block] == NULL && q->binds[block] == NULL) {
                     /* if there's nothing special about this block, merge up */
                     mergable = 1;
-                } else if (q->constraints[parent] == NULL && q->blocks[parent].length == 0) {
+                } else if (q->constraints[parent] == NULL && q->binds[block] == NULL && q->blocks[parent].length == 0) {
                     mergable = 1;
                 }
             }
@@ -314,6 +313,10 @@ static void tree_compact(fs_query *q)
                     q->constraints[parent] = q->constraints[block];
                     q->constraints[block] = NULL;
                 }
+                if (q->binds[parent] == NULL) {
+                    q->binds[parent] = q->binds[block];
+                    q->binds[block] = NULL;
+                }
                 for (int col=0; q->bb[0][col].name; col++) {
                     if (q->bb[0][col].appears == block) {
                         q->bb[0][col].appears = parent;
@@ -331,11 +334,10 @@ static void tree_compact(fs_query *q)
     } while (done_something);
 }
 
-fs_query *fs_query_execute(fs_query_state *qs, fsp_link *link, raptor_uri *bu, const char *query, unsigned int flags, int opt_level, int soft_limit, int explain)
+fs_query *fs_query_execute(fs_query_state *qs, fsp_link *link, raptor_uri *bu, const char *query, unsigned int flags, int opt_level, int soft_limit,const char *apikey, int explain)
 {
     if (!qs) {
         fs_error(LOG_CRIT, "fs_query_execute() handed NULL query state");
-
         return NULL;
     }
 
@@ -363,6 +365,12 @@ fs_query *fs_query_execute(fs_query_state *qs, fsp_link *link, raptor_uri *bu, c
     q->rq = rq;
     q->qs = qs;
     q->opt_level = opt_level;
+
+    if (fsp_is_acl_enabled(qs->link) && apikey)
+        q->apikey_rid = fs_hash_literal(apikey,0);
+    else
+        q->apikey_rid = FS_RID_NULL;
+
     if (soft_limit) {
         q->soft_limit = soft_limit;
     } else {
@@ -420,6 +428,8 @@ fs_query *fs_query_execute(fs_query_state *qs, fsp_link *link, raptor_uri *bu, c
         q->default_graphs->data[0] = FS_DEFAULT_GRAPH_RID;
     }
 
+    q->tmp_resources = g_hash_table_new_full(fs_rid_hash, fs_rid_equal, g_free, fs_free_cached_resource);
+
 #ifdef DEBUG_MERGE
     explain = 1;
     rasqal_query_print(rq, stdout);
@@ -460,7 +470,9 @@ fs_query *fs_query_execute(fs_query_state *qs, fsp_link *link, raptor_uri *bu, c
             q->default_graphs->length = 0;
         }
         char *uri = (char *)raptor_uri_as_string(dg->uri);
-        char *name_uri = (char *)raptor_uri_as_string(dg->name_uri);
+        char *name_uri = NULL;
+        if (dg->name_uri)
+            name_uri = (char *)raptor_uri_as_string(dg->name_uri);
         if (name_uri) {
             q->warnings = g_slist_prepend(q->warnings, "FROM NAMED is not currently supported");
         } else {
@@ -521,9 +533,14 @@ fs_query *fs_query_execute(fs_query_state *qs, fsp_link *link, raptor_uri *bu, c
     if (rasqal_query_get_group_condition(rq, 0) ||
         rasqal_query_get_having_condition(rq, 0)) {
         q->aggregate = 1;
+        /* if aggregate then offset must be evaluated in the result generation */
+        if (q->aggregate) {
+            q->offset_aggregate = q->offset;
+            q->offset = 0;
+        }
     }
 #if 0
-// This nees to be be refined so it only applies ot queries without FILTERs etc.
+// This needs to be be refined so it only applies to queries without FILTERs etc.
  else {
         if (!rasqal_query_get_distinct(rq) && q->limit) {
             q->soft_limit = q->limit * 100;
@@ -777,7 +794,7 @@ int fs_query_process_pattern(fs_query *q, rasqal_graph_pattern *pattern, raptor_
 	    /* execute triple pattern query */
 	    if (explain) {
                 FILE *msg = tmpfile();
-		fprintf(msg, "execute: ");
+		fprintf(msg, "execute %d/%d: ", j, q->blocks[i].length);
 		if (!q->blocks[i].data[j]) {
 		    fprintf(msg, "NULL");
 		} else {
@@ -842,6 +859,34 @@ int fs_query_process_pattern(fs_query *q, rasqal_graph_pattern *pattern, raptor_
                 break;
             }
 	}
+
+        /* Evaluate BIND expressions */
+        fs_bind_expression *be = q->binds[i];
+        const int bind_length = fs_binding_length(q->bb[i]);
+        while (be) {
+            fs_binding *beb = fs_binding_get(q->bb[i], be->var);
+            // This is needed for fs_expression_eval() - don't ask
+            q->bt = q->bb[i];
+            beb->bound = 1;
+            if (beb->vals->length < bind_length) {
+                fs_rid_vector_grow(beb->vals, bind_length);
+            }
+            for (int row=0; row < bind_length; row++) {
+                fs_value eval = fs_expression_eval(q, row, i, be->expr);
+                eval = fs_value_fill_lexical(q, eval);
+                eval = fs_value_fill_rid(q, eval);
+                /* insert entry if it's new */
+                if (!g_hash_table_lookup(q->tmp_resources, &eval.rid)) {
+                    fs_resource *res = fs_resource_value(q, eval);
+                    fs_rid *re = malloc(sizeof(fs_rid));
+                    *re = eval.rid;
+                    g_hash_table_insert(q->tmp_resources, re, res);
+                }
+                beb->vals->data[row] = eval.rid;
+            }
+            be = be->next;
+        }
+
 #if DEBUG_MERGE > 1
         printf("table after processing B%d:\n", i);
         fs_binding_print(q->bb[i], stdout);
@@ -899,7 +944,7 @@ int fs_query_process_pattern(fs_query *q, rasqal_graph_pattern *pattern, raptor_
                     printf("block B%d is in UNION group %d\n", j, q->union_group[j]);
 #endif
                     /* It's a UNION */
-                    /* apply constriants now, it's too trick to delay execution */
+                    /* apply constriants now, it's too tricky to delay execution */
                     if (q->constraints[j]) {
                         fs_binding *old = q->bb[j];
                         q->bb[j] = fs_binding_apply_filters(q, j, q->bb[j], q->constraints[j]);
@@ -1013,6 +1058,28 @@ void fs_query_free(fs_query *q)
 
         if (q->default_graphs) fs_rid_vector_free(q->default_graphs);
 
+        for (int i=0;i<FS_MAX_BLOCKS;i++) {
+            if (q->constraints[i]) {
+                raptor_free_sequence(q->constraints[i]);
+            }
+            if (q->binds[i]) {
+                fs_bind_expression *be = q->binds[i];
+                while (be) {
+                    fs_bind_expression *next = be->next;
+                    free(be);
+                    be = next;
+                }
+            }
+        }
+
+        /* If the query object was created unconventionally it might not exist */
+        if (q->tmp_resources) {
+            g_hash_table_destroy(q->tmp_resources);
+        }
+        /* we forgot to do this somewhere else */
+        if (q->free_row_list)
+            fs_query_free_row_freeable(q);
+
         memset(q, 0, sizeof(fs_query));
 	free(q);
     }
@@ -1164,9 +1231,8 @@ static void filter_optimise_disjunct_equality(fs_query *q,
         if (val == FS_RID_NULL) return;
         /* if it's a value that can have multiple lexical forms for one
            conceptual value, then we need to check it with a real FILTER */
-        if (attr == fs_c.xsd_integer || attr == fs_c.xsd_double ||
-            attr == fs_c.xsd_float || attr == fs_c.xsd_decimal ||
-            attr == fs_c.xsd_string || attr == fs_c.xsd_datetime) {
+        /* Things that are OK are simple literals, URIs, and plain literals */
+        if (attr != FS_RID_NULL && attr != fs_c.empty && !FS_IS_LITERAL(attr)) {
             fs_rid_vector_truncate(res, 0);
 
             return;
@@ -1193,12 +1259,18 @@ static void filter_optimise_disjunct_equality(fs_query *q,
         
     if (e->arg1) {
         filter_optimise_disjunct_equality(q, e->arg1, block, var, res);
+
+        if (res->length == 0) return;
     }
     if (e->arg2) {
         filter_optimise_disjunct_equality(q, e->arg2, block, var, res);
+
+        if (res->length == 0) return;
     }
     if (e->arg3) {
         filter_optimise_disjunct_equality(q, e->arg3, block, var, res);
+
+        if (res->length == 0) return;
     }
 }
 
@@ -1268,8 +1340,8 @@ static void graph_pattern_walk(fsp_link *link, rasqal_graph_pattern *pattern,
         rasqal_expression *e =
             rasqal_graph_pattern_get_filter_expression(pattern);
         if (e) {
-            /* we need to flag if it's a UNION FILTER so we don't
-             * unneccesarily set the selected flag on the variables */
+            /* we need to flag if it's a UNION FILTER (FILTER disjunction) so
+             * we don't unneccesarily set the selected flag on the variables */
             check_variables(q, e, uni);
             if (filter_optimise(q, e, q->block)) {
                 /* stop us from trying to evaluate this expression later */
@@ -1288,15 +1360,32 @@ static void graph_pattern_walk(fsp_link *link, rasqal_graph_pattern *pattern,
             raptor_sequence_push(q->constraints[parent], e);
         }
         break;
+    case RASQAL_GRAPH_PATTERN_OPERATOR_LET:
+        handled = 1;
+        rasqal_variable *v = rasqal_graph_pattern_get_variable(pattern);
+	fs_binding *b = fs_binding_get(q->bb[0], v);
+	if (b) {
+            b->need_val = 1;
+        }
+        fs_bind_expression *be = malloc(sizeof(fs_bind_expression));
+        be->var = v;
+        be->expr = rasqal_graph_pattern_get_filter_expression(pattern);
+        be->next = q->binds[q->block];
+        q->binds[q->block] = be;
+        check_variables(q, be->expr, 0);
+        break;
     case RASQAL_GRAPH_PATTERN_OPERATOR_MINUS:
         handled = 1;
 	(q->block)++;
         q->parent_block[q->block] = parent;
         q->join_type[q->block] = FS_MINUS;
         break;
+    case RASQAL_GRAPH_PATTERN_OPERATOR_SELECT:
+        q->errors++;
+        q->warnings = g_slist_prepend(q->warnings, "SubSELECTs are not implemented");
     }
     if (!handled) {
-	fs_error(LOG_ERR, "Unknown GP operator %d not supported", op);
+	fs_error(LOG_ERR, "Unknown GP operator %s (%d) not supported", rasqal_graph_pattern_operator_as_string(op), op);
     }
 
     if (uni) {
@@ -1638,6 +1727,7 @@ static int process_results(fs_query *q, int block, fs_binding *oldb,
                 fs_binding *bv = fs_binding_get(b, vars[col]);
                 if (!bv) {
                     fs_error(LOG_CRIT, "unmatched variable name '%s'", vars[col]->name);
+                    fs_rid_vector_free(results[col]);
                     continue;
                 }
                 /* if the varaible is repeated in one triple pattern then it
@@ -1651,7 +1741,10 @@ static int process_results(fs_query *q, int block, fs_binding *oldb,
                             break;
                         }
                     }
-                    if (repeat) continue;
+                    if (repeat) {
+                        fs_rid_vector_free(results[col]);
+                        continue;
+                    }
                 }
 		fs_binding_add_vector(b, vars[col], results[col]);
                 ret += results[col] ? results[col]->length : 0;
@@ -1910,6 +2003,8 @@ static fs_rid const_literal_to_rid(fs_query *q, rasqal_literal *l, fs_rid *attr)
     switch (l->type) {
 	case RASQAL_LITERAL_URI: {
             char *uri = (char *)raptor_uri_as_string(l->value.uri);
+            *attr = FS_RID_NULL;
+
             return fs_hash_uri(uri);
         }
         case RASQAL_LITERAL_XSD_STRING:
@@ -1923,30 +2018,44 @@ static fs_rid const_literal_to_rid(fs_query *q, rasqal_literal *l, fs_rid *attr)
             } else if (l->datatype) {
                 *attr = fs_hash_uri((char *)raptor_uri_as_string(l->datatype));
             }
+
 	    return fs_hash_literal((char *)l->string, *attr);
         }
 	case RASQAL_LITERAL_BOOLEAN:
             *attr = fs_c.xsd_boolean;
+
 	    return fs_hash_literal(l->value.integer ?
 			"true" : "false", *attr);
 	case RASQAL_LITERAL_INTEGER:
-        *attr = fs_c.xsd_integer;
+            *attr = fs_c.xsd_integer;
+
 	    return fs_hash_literal((char *)l->string, *attr);
 	case RASQAL_LITERAL_INTEGER_SUBTYPE:
             *attr = fs_c.xsd_int;
+
 	    return fs_hash_literal((char *)l->string, *attr);
 	case RASQAL_LITERAL_DOUBLE:
             *attr = fs_c.xsd_double;
+
 	    return fs_hash_literal((char *)l->string, *attr);
 	case RASQAL_LITERAL_FLOAT:
             *attr = fs_c.xsd_float;
+
 	    return fs_hash_literal((char *)l->string, *attr);
 	case RASQAL_LITERAL_DECIMAL:
             *attr = fs_c.xsd_decimal;
+
 	    return fs_hash_literal((char *)l->string, *attr);
 	case RASQAL_LITERAL_DATETIME:
             *attr = fs_c.xsd_datetime;
+
 	    return fs_hash_literal((char *)l->string, *attr);
+#if RASQAL_VERSION >= 929
+	case RASQAL_LITERAL_DATE:
+            *attr = fs_c.xsd_date;
+
+	    return fs_hash_literal((char *)l->string, *attr);
+#endif
 	case RASQAL_LITERAL_VARIABLE:
             /* not const, don't handle here */
             break;
@@ -2076,6 +2185,14 @@ int fs_bind_slot(fs_query *q, int block, fs_binding *b,
 	    }
 	    fs_rid_vector_append(v, fs_hash_literal((char *)l->string, fs_c.xsd_datetime));
 	    break;
+#if RASQAL_VERSION >= 929
+	case RASQAL_LITERAL_DATE:
+	    if (!lit_allowed) {
+		return 1;
+	    }
+	    fs_rid_vector_append(v, fs_hash_literal((char *)l->string, fs_c.xsd_date));
+	    break;
+#endif
 	case RASQAL_LITERAL_BLANK:
 	case RASQAL_LITERAL_PATTERN:
 	case RASQAL_LITERAL_QNAME:
