@@ -68,6 +68,8 @@ static long all_time_import_count = 0;
 static int global_import_count = 0;
 static int unsafe = 0;
 static int default_graph = 0;
+static int cache_stats = 0;
+static int graph_access_control = 0;
 static int soft_limit = 0; /* default value for soft limit */
 static int opt_level = -1;  /* default value for optimisation level */
 static int cors_support = -1; /* cross-origin resource sharing (CORS) support */
@@ -91,6 +93,13 @@ volatile static unsigned int last_query_id = 0;
 static void set_boolean(GKeyFile *keyfile, const char *kb_name, const char *key, int *set);
 static void set_string(GKeyFile *keyfile, const char *kb_name, const char *key, const char **set);
 
+static void fs_free_global_elements() {
+  if (fsplink) {
+    fsp_free_acl_system(fsplink);
+    fsp_close_link(fsplink);
+  }
+}
+
 static void query_log_open (const char *kb_name)
 {
   char *filename = g_strdup_printf("/var/log/4store/query-%s.log", kb_name);
@@ -126,8 +135,10 @@ static void query_log (client_ctxt *ctxt, const char *query)
     struct tm *ts = gmtime(&now);
     char time_str[21];
     strftime(time_str, sizeof(time_str), "%Y-%m-%dT%H:%M:%SZ", ts);
-
-    fprintf(ql_file, "##### %s Q%u\n%s\n", time_str, ctxt->query_id, query);
+    if (!ctxt->apikey)
+      fprintf(ql_file, "##### %s Q%u-pid%u\n%s\n", time_str, ctxt->query_id, cpid, query);
+    else
+      fprintf(ql_file, "##### %s Q%u-pid%u %s\n%s\n", time_str, ctxt->query_id, cpid, ctxt->apikey, query);
     fflush(ql_file);
   }
 }
@@ -318,6 +329,8 @@ static void client_free(client_ctxt *ctxt)
   }
   g_hash_table_destroy(ctxt->headers);
   free(ctxt->request);
+  if (ctxt->apikey)
+    g_free(ctxt->apikey);
   g_free(ctxt);
 }
 
@@ -336,7 +349,9 @@ static void http_query_worker(gpointer data, gpointer user_data)
   client_ctxt *ctxt = (client_ctxt *) data;
 
   ctxt->start_time = fs_time();
-  ctxt->qr = fs_query_execute(query_state, fsplink, bu, ctxt->query_string, ctxt->query_flags, opt_level, ctxt->soft_limit, 0);
+  ctxt->qr = fs_query_execute(query_state, fsplink, bu, ctxt->query_string, 
+                              ctxt->query_flags, opt_level, ctxt->soft_limit, 
+                              ctxt->apikey, 0);
   if (ctxt->qr->errors) {
     http_error(ctxt, "400 Parser error");
     GSList *w = ctxt->qr->warnings;
@@ -381,11 +396,11 @@ static void http_query_worker(gpointer data, gpointer user_data)
 
     if (ctxt->output) {
       type = ctxt->output;
-    } else if (ctxt->qr->construct && accept && strstr(accept, "text/turtle")) {
+    } else if ((ctxt->qr->construct || ctxt->qr->describe) && accept && strstr(accept, "text/turtle")) {
       type = "text";
       fprintf(fp, "Content-Type: text/turtle\r\n\r\n");
       flags = 0;
-    } else if (ctxt->qr->construct && accept && strstr(accept, "application/rdf+xml")) {
+    } else if ((ctxt->qr->construct || ctxt->qr->describe) && accept && strstr(accept, "application/rdf+xml")) {
       type = "sparql";
     } else if (accept && strstr(accept, "application/sparql-results+xml")) {
       type = "sparql";
@@ -451,6 +466,18 @@ static gboolean import_watchdog (gpointer data)
   return FALSE;
 }
 
+static int data_modification_acl_granted(client_ctxt *ctxt) {
+    if (fsp_is_acl_enabled(fsplink)) {
+        if (fsp_acl_needs_reload(fsplink))
+          fs_acl_load_system_info(fsplink);
+        if (!ctxt->apikey ||
+           !fsp_acl_is_admin(fsplink, fs_hash_literal(ctxt->apikey,0))) {
+              return 0;
+        }
+    }
+    return 1; 
+}
+
 static void http_import_start(client_ctxt *ctxt)
 {
   /* If it's an update operation, we have a different path */
@@ -634,9 +661,45 @@ static void http_put_finished(client_ctxt *ctxt, const char *msg)
 
 static void http_put_request(client_ctxt *ctxt, gchar *url, gchar *protocol)
 {
-  if (!strncmp(url, "/data/?graph=", 13)) { /* SPARQL 1.1 way */
-    url += 13;
-    url_decode(url);
+
+  int sparql11 = !strncmp(url, "/data/?", 7);  /* SPARQL 1.1 way */
+  /* access control only implmented in SPARQL 1.1 way */
+
+  if (sparql11) {
+    char *qm = strchr(url, '?');
+    char *qs = qm ? qm + 1 : NULL;
+    if (qs) {
+      *qm = '\0';
+    }
+    char *path = url;
+    url_decode(path);
+    while (qs) {
+      char *ampersand = strchr(qs, '&');
+      char *next = ampersand ? ampersand + 1 : NULL;
+      if (next) {
+        *ampersand = '\0';
+      }
+      char *key = qs;
+      char *equals = strchr(qs, '=');
+      char *value = equals ? equals + 1 : NULL;
+      if (equals) {
+        *equals = '\0';
+      }
+      url_decode(key);
+      if (!strcmp(key, "graph") && value) {
+        url_decode(value);
+        url = value;
+      } else if (!strcmp(key, "apikey") && value) {
+        url_decode(value);
+        ctxt->apikey = g_strdup(value);
+      }
+      qs = next;
+    }
+    if (!data_modification_acl_granted(ctxt)) {
+        http_error(ctxt, "403 forbidden - updates only with admin API KEY when ACL is enabled");
+        http_close(ctxt);
+        return;
+    }
   } else if (!strncmp(url, "/data/", 6)) { /* pre-SPARQL 1.1 4store way */
     url += 6;
     url_decode(url);
@@ -666,10 +729,51 @@ static void http_put_request(client_ctxt *ctxt, gchar *url, gchar *protocol)
 
 static void http_delete_request(client_ctxt *ctxt, gchar *url, gchar *protocol)
 {
-  if (!strncmp(url, "/data/?graph=", 13)) { /* SPARQL 1.1 way */
-    url += 13;
-    url_decode(url);
+
+  int sparql11 = !strncmp(url, "/data/?", 7);  /* SPARQL 1.1 way */
+  /* access control only implmented in SPARQL 1.1 way */
+
+  if (sparql11) {
+    char *qm = strchr(url, '?');
+    char *qs = qm ? qm + 1 : NULL;
+    if (qs) {
+      *qm = '\0';
+    }
+    char *path = url;
+    url_decode(path);
+    while (qs) {
+      char *ampersand = strchr(qs, '&');
+      char *next = ampersand ? ampersand + 1 : NULL;
+      if (next) {
+        *ampersand = '\0';
+      }
+      char *key = qs;
+      char *equals = strchr(qs, '=');
+      char *value = equals ? equals + 1 : NULL;
+      if (equals) {
+        *equals = '\0';
+      }
+      url_decode(key);
+      if (!strcmp(key, "graph") && value) {
+        url_decode(value);
+        url = value;
+      } else if (!strcmp(key, "apikey") && value) {
+        url_decode(value);
+        ctxt->apikey = g_strdup(value);
+      }
+      qs = next;
+    }
+    if (!data_modification_acl_granted(ctxt)) {
+        http_error(ctxt, "403 forbidden - updates only with admin API KEY when ACL is enabled");
+        http_close(ctxt);
+        return;
+    }
   } else if (!strncmp(url, "/data/", 6)) { /* pre-SPARQL 1.1 4store syntax */
+    #if 0 /* uncomment if only SPARQL 1.1 way with access control */
+        http_error(ctxt, "403 forbidden - updates only with admin API KEY when ACL is enabled");
+        http_close(ctxt);
+        return;
+    #endif 
     url += 6;
     url_decode(url);
   } else if (!strncmp(url, "/", 1)) {
@@ -750,6 +854,83 @@ static void http_status_report(client_ctxt *ctxt)
   http_close(ctxt);
 }
 
+static void http_cache_report(client_ctxt *ctxt) {
+
+  http_send(ctxt, "HTTP/1.0 200 OK\r\n"
+  "Server: 4s-httpd/" GIT_REV "\r\n"
+  "Content-Type: text/html; charset=UTF-8\r\n"
+  "\r\n"
+  "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\n"
+  "<html><head><title>SPARQL httpd server status - cache</title></head>\n"
+  "<body><h1>SPARQL httpd server " GIT_REV " status - cache</h1>\n"
+  "<h2>KB ");
+  http_send(ctxt, fsp_kb_name(fsplink));
+  http_send(ctxt, "</h2>\n");
+  http_send(ctxt, "<h3>RID cache stats</h3>\n");
+  http_send(ctxt, "<table border=1 cellpadding=6>\n");
+
+  g_static_mutex_lock(&query_state->cache_mutex);
+  char *line = g_strdup_printf("<tr><td>cache_hits</td><td>%d</td></tr>\n",query_state->cache_hits);
+  http_send(ctxt, line);
+  g_free(line);
+  
+  line = g_strdup_printf("<tr><td>resolve_all_calls</td><td>%d (avg elapse %.4f ms)</td></tr>\n",
+    query_state->resolve_all_calls,query_state->resolve_all_elapse*1000.0);
+  http_send(ctxt, line);
+  g_free(line);
+
+  line = g_strdup_printf("<tr><td>avg_cache_saves_l2</td><td>%.4f%%</td></tr>\n",
+    query_state->avg_cache_saves_l2 * 100);
+  http_send(ctxt, line);
+  g_free(line);
+
+  line = g_strdup_printf("<tr><td>cache_success_l2</td><td>%d (%.2f%%)</td></tr>\n",
+    query_state->cache_success_l2,
+    100 * (query_state->cache_success_l2 / (query_state->cache_hits+0.000)));
+  http_send(ctxt, line);
+  g_free(line);
+  
+  line = g_strdup_printf("<tr><td>cache_success_l1</td><td>%d (%.2f%%)</td></tr>\n",
+    query_state->cache_success_l1, 
+    100 * (query_state->cache_success_l1 / (query_state->cache_hits+0.000)));
+  http_send(ctxt, line);
+  g_free(line);
+
+  line = g_strdup_printf("<tr><td>cache_fail</td><td>%d (%.2f%%)</td></tr>\n",
+    query_state->cache_fail,
+    100 * (query_state->cache_fail / (query_state->cache_hits+0.000)));
+  http_send(ctxt, line);
+  g_free(line);
+  http_send(ctxt, "</table>\n");
+
+  http_send(ctxt, "<h3>BIND cache stats</h3>\n");
+  http_send(ctxt, "<table border=1 cellpadding=6>\n");
+
+  line = g_strdup_printf("<tr><td>bind_hits</td><td>%d</td></tr>\n",
+    query_state->bind_hits);
+  http_send(ctxt, line);
+  g_free(line);
+
+  line = g_strdup_printf("<tr><td>bind_cache_success</td><td>%d (%.4f%%)</td></tr>\n",
+    query_state->bind_cache_success,
+    100 * (query_state->bind_cache_success / (query_state->bind_hits+0.000)));
+  http_send(ctxt, line);
+  g_free(line);
+
+  unsigned int count_bind = fs_query_bind_cache_count_slots(query_state);
+  line = g_strdup_printf("<tr><td>bind_slots</td><td>%d (%.4f%%)</td></tr>\n",
+    count_bind,
+    100 * (count_bind / (fs_query_bind_cache_size()+0.0001)));
+  http_send(ctxt, line);
+  g_free(line);
+  
+  g_static_mutex_unlock(&query_state->cache_mutex);
+  http_send(ctxt, "</table>\n");
+
+  http_send(ctxt, "</body></html>\n");
+  http_close(ctxt);
+}
+
 static void http_size_report(client_ctxt *ctxt)
 {
   http_send(ctxt, "HTTP/1.0 200 OK\r\n"
@@ -798,7 +979,7 @@ static void http_service_description(client_ctxt *ctxt)
 {
   http_send(ctxt, "HTTP/1.0 200 OK\r\n"
   "Server: 4s-httpd/" GIT_REV "\r\n"
-  "Content-Type: application/x-turtle\r\n"
+  "Content-Type: text/turtle; charset=utf-8\r\n"
   "\r\n"
   "@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .\n"
   "@prefix sd: <http://darq.sf.net/dose/0.1#> .\n"
@@ -933,10 +1114,17 @@ static void http_get_request(client_ctxt *ctxt, gchar *url, gchar *protocol)
       } else if (!strcmp(key, "default-graph-uri") && value) {
         url_decode(value);
         default_graph = value;
+      } else if (!strcmp(key, "apikey") && value) {
+        url_decode(value);
+        ctxt->apikey = g_strdup(value);
       }
       qs = next;
     }
-    if (query) {
+    if (graph_access_control && !ctxt->apikey) {
+        http_error(ctxt, "403 forbidden - apikey parameter has to be included in request.");
+        http_close(ctxt);
+    }
+    else if (query) {
       http_answer_query(ctxt, query);
     } else {
       http_error(ctxt, "500 SPARQL protocol error");
@@ -953,6 +1141,9 @@ static void http_get_request(client_ctxt *ctxt, gchar *url, gchar *protocol)
   } else if (!strcmp(path, "/status")) {
     http_redirect(ctxt, "status/");
     http_close(ctxt);
+  } else if (!strcmp(path, "/status/cache/") ||
+             !strcmp(path, "/status/cache")) {
+    http_cache_report(ctxt);
   } else if (!strcmp(path, "/status/size/")) {
     http_size_report(ctxt);
   } else if (!strcmp(path, "/description/")) {
@@ -1062,10 +1253,17 @@ static void http_post_request(client_ctxt *ctxt, gchar *url, gchar *protocol)
       } else if (!strcmp(key, "default-graph-uri") && value) {
         url_decode(value);
         default_graph = value;
+      } else if (!strcmp(key, "apikey") && value) {
+        url_decode(value);
+        ctxt->apikey = g_strdup(value);
       }
       qs = next;
     }
-    if (query) {
+    if (graph_access_control && !ctxt->apikey) {
+        http_error(ctxt, "403 forbidden - apikey parameter has to be included in request.");
+        http_close(ctxt);
+    }
+    else if (query) {
       http_answer_query(ctxt, query);
     } else {
       http_error(ctxt, "500 SPARQL protocol error");
@@ -1126,17 +1324,25 @@ static void http_post_request(client_ctxt *ctxt, gchar *url, gchar *protocol)
       if (!strcmp(key, "update") && value) {
         url_decode(value);
         update = value;
+      } else if (!strcmp(key, "apikey") && value) {
+        url_decode(value);
+        ctxt->apikey = g_strdup(value);
       }
       qs = next;
     }
     if (update) {
-      ctxt->update_string = update;
-      g_source_remove_by_user_data(ctxt);
-      if (import_queue) {
-        import_queue = g_slist_append(import_queue, ctxt);
+      if (!data_modification_acl_granted(ctxt)) {
+        http_error(ctxt, "403 forbidden - updates only with admin API KEY when ACL is enabled");
+        http_close(ctxt);
       } else {
-        import_queue = g_slist_append(import_queue, ctxt);
-        http_import_start(ctxt);
+        ctxt->update_string = update;
+        g_source_remove_by_user_data(ctxt);
+        if (import_queue) {
+          import_queue = g_slist_append(import_queue, ctxt);
+        } else {
+          import_queue = g_slist_append(import_queue, ctxt);
+          http_import_start(ctxt);
+        }
       }
     } else {
       http_error(ctxt, "500 SPARQL protocol error");
@@ -1207,11 +1413,20 @@ static void http_post_request(client_ctxt *ctxt, gchar *url, gchar *protocol)
       } else if (!strcmp(key, "mime-type") && value) {
         url_decode(value);
         mime_type = value;
+      } else if (!strcmp(key, "apikey") && value) {
+        url_decode(value);
+        ctxt->apikey = g_strdup(value);
       }
       qs = next;
     }
+
     if (graph && data) {
-      http_post_data(ctxt, graph, mime_type, data);
+      if (!data_modification_acl_granted(ctxt)) {
+            http_error(ctxt, "403 forbidden - updates only with admin API KEY when ACL is enabled");
+            http_close(ctxt);
+      } else {
+        http_post_data(ctxt, graph, mime_type, data);
+      }
     } else {
       http_error(ctxt, "500 SPARQL REST protocol error");
       http_close(ctxt);
@@ -1633,6 +1848,11 @@ static int server_setup (int background, const char *host, const char *port)
   } else {
     fs_error(LOG_INFO, "4store HTTP daemon " GIT_REV " started on port %s", port);
   }
+  if (cache_stats)
+    fs_error(LOG_ERR,"Cache stats enabled at /status/cache.");
+  if (graph_access_control)
+    fs_error(LOG_ERR,"Access control for graphs is enabled.");
+  
   return srv;
 }
 
@@ -1658,6 +1878,9 @@ static void child (int srv, char *kb_name, char *password)
   query_log_open(kb_name);
 
   query_state = fs_query_init(fsplink, NULL, NULL);
+  query_state->cache_stats = cache_stats;
+  if (graph_access_control)
+    fsp_init_acl_system(fsplink);
   bu = raptor_new_uri(query_state->raptor_world, (unsigned char *)"local:local");
   g_thread_init(NULL);
   pool = g_thread_pool_new(http_query_worker, NULL, QUERY_THREAD_POOL_SIZE, FALSE, NULL);
@@ -1696,6 +1919,7 @@ static void child_exited(pid_t pid, gint status)
     int code = WTERMSIG(status);
     fs_error((code == SIGTERM || code == SIGKILL) ? LOG_INFO : LOG_CRIT,
              "child %d terminated by signal %d", pid, code);
+    fs_free_global_elements();
   } else if (WIFSTOPPED(status)) {
     fs_error(LOG_ERR, "child %d stopped by signal %d", pid, WSTOPSIG(status));
   } else {
@@ -1721,7 +1945,7 @@ int main(int argc, char *argv[])
 
 
   int o;
-  while (!help && (o = getopt(argc, argv, "DH:p:Uds:O:X")) != -1) {
+  while (!help && (o = getopt(argc, argv, "DCAH:p:Uds:O:Xc:")) != -1) {
     switch (o) {
       case 'D':
         daemonize = 0;
@@ -1733,32 +1957,41 @@ int main(int argc, char *argv[])
         port = optarg;
         break;
       case 'U':
-	unsafe = 1;
-	break;
+        unsafe = 1;
+        break;
       case 'd':
-	default_graph = 1;
-	break;
+        default_graph = 1;
+        break;
       case 's':
-	soft_limit = atoi(optarg);
-	if (soft_limit == 0) {
-	  /* -1 means off */
-	  soft_limit = -1;
-	}
-	break;
+        soft_limit = atoi(optarg);
+        if (soft_limit == 0) {
+          /* -1 means off */
+          soft_limit = -1;
+        }
+        break;
       case 'O':
-	opt_level = atoi(optarg);
-	break;
+        opt_level = atoi(optarg);
+        break;
       case 'X':
-	cors_support = 1;
-	break;
+        cors_support = 1;
+        break;
+      case 'c':
+        fs_set_config_file(optarg);
+        break;
+      case 'C':
+        cache_stats = 1;
+        break;
+      case 'A':
+        graph_access_control = 1;
+        break;
       default:
-	help = 1;
-	break;
+        help = 1;
+        break;
     }
   }
 
   if (help || optind >= argc || optind < argc - 1) {
-    fprintf(stdout, "Usage: %s [-D] [-H host] [-p port] [-U] [-s limit] <kbname>\n", basename(argv[0]));
+    fprintf(stdout, "Usage: %s [-D] [-H host] [-p port] [-U] [-s limit] [-c path] <kbname>\n", basename(argv[0]));
     fprintf(stdout, "       -H   specify host to listen on\n");
     fprintf(stdout, "       -p   specify port to listen on\n");
     fprintf(stdout, "       -D   do not daemonise\n");
@@ -1767,6 +2000,9 @@ int main(int argc, char *argv[])
     fprintf(stdout, "       -s   default soft limit (-1 to disable)\n");
     fprintf(stdout, "       -O   set query optimiser level (0-3, default is 3)\n");
     fprintf(stdout, "       -X   enable public cross-origin resource sharing (CORS) support\n");
+    fprintf(stdout, "       -c   path to config file\n");
+    fprintf(stdout, "       -C   enable cache stats in /status/cache\n");
+    fprintf(stdout, "       -A   enable access control at graph level\n");
     fprintf(stdout, "Options can also be set permenantly in /etc/4store.conf\n");
     fprintf(stdout, "see http://4store.org/trac/wiki/SparqlServer for details\n");
 
@@ -1780,7 +2016,7 @@ int main(int argc, char *argv[])
 
   GKeyFile *keyfile = g_key_file_new();
   GError *err = NULL;
-  const char *keyfile_filename = FS_CONFIG_FILE;
+  const char *keyfile_filename = fs_get_config_file();
   if (!g_key_file_load_from_file(keyfile, keyfile_filename, G_KEY_FILE_KEEP_COMMENTS, &err)) {
     if (err->code != G_FILE_ERROR_NOENT &&
         err->code != G_FILE_ERROR_EXIST &&
@@ -1802,6 +2038,10 @@ int main(int argc, char *argv[])
 
     set_string(keyfile, kb_name, "listen", &host);
 
+    set_boolean(keyfile, kb_name, "cache-stats", &cache_stats);
+    
+    set_boolean(keyfile, kb_name, "graph-access-control", &graph_access_control);
+
     if (soft_limit == 0) {
       const char *soft_limit_str = NULL;
       set_string(keyfile, kb_name, "soft-limit", &soft_limit_str);
@@ -1821,6 +2061,7 @@ int main(int argc, char *argv[])
       }
     }
   }
+  g_key_file_free(keyfile);
 
   /* handle defaults */
 
